package org.coderu.common.utils;


import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.lang3.Validate;
import org.coderu.common.utils.types.Pair;

import com.google.common.base.Function;
import com.google.common.base.Functions;
import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multisets;



/**
 * A set of collection related utility methods.
 * Partly implemented as a generic wrapper to CollectionUtils from apache
 *
 * @author Alexander Rubinov
 *
 */
public final class CollectionUtils {


    public static <F, T> Collection<T> transform(
        final Collection<F> fromCollection,Function<? super F, T> function) {
      return new ArrayList<T>(Collections2.transform(fromCollection, function));
    }
    
    public static <E> Collection<E> filter(
        Collection<E> unfiltered, Predicate<? super E> predicate) {
        return new ArrayList<E>(Collections2.filter(unfiltered, predicate));
      }
    
    /**
     * Returns a Collection containing all elements from both given Collections.
     * 
     * The order of elements will be retain.
     * 
     * @param <T> type of elements of collection
     * @param a first collection
     * @param b second collections
     * @return collection containing the union of the given Collections.
     */
    public static <T> List<T> concat( final Collection<? extends T> a, Collection<? extends T> b){
        final List<T> result = new ArrayList<T>(a.size()+b.size());
        result.addAll(a);
        result.addAll(b);
        return result;
    }

    /**
     * Returns a Collection containing all elements from all given Collections.
     * 
     * The order of elements will be retain.
     * 
     * @param <T> type of elements of collection
     * @param a the list of collections to union
     * @return a collection containing the union of the given collections.
     */
    @SuppressWarnings("unchecked")
    public static <T> Collection<T> concat( final Collection<? extends Collection<T>> a){
        Collection<T> result = Collections.EMPTY_LIST;

        for(Iterator<? extends Collection<T>> iter = a.iterator();iter.hasNext();){
            result = concat(result,iter.next());
        }

        return result;
    }
    
    /**
     * Returns a Collection containing all elements from all given Collections.
     * 
     * The order of elements will be retain.
     * 
     * @param <T> type of elements of collection
     * @param a the list of collections to union
     * @return a collection containing the union of the given collections.
     */
    public static <T> Collection<T> concat( final Collection<T>... a){
        return concat(Arrays.asList(a));
    }

    /**
     * gets a collection containing a - b.
     * The cardinality of each element e in the returned Collection  will be the cardinality of
     * e in a minus the cardinality of e in b, or zero, whichever is greater.
     * @return a collection containing a - b
     */
    public static <T> Collection<T> subtract( final Collection<T> a, final Collection<T> b){
        
        final Multiset<T> multisetToModify = HashMultiset.create(a);
        final Multiset<T> occurrencesToRemove = HashMultiset.create(b);
        Multisets.removeOccurrences(multisetToModify, occurrencesToRemove);
        return multisetToModify;
    }

    /**
     * converts a singleton collection to first (and sole) element.
     * @param <T> type of collection
     * @param collection the collection
     * @exception AssertionViolation if collection does't contain exactly one element
     * @return the first (and sole) element
     */
    public static <T> T col2one(final Collection<T> collection){
        Validate.isTrue(collection.size()==1, "size is not equal one");
        return collection.iterator().next();
    }

    

    /**
     * get first element of collection that match predicate or null otherwise
     * @param <T> type of elements of collection
     * @param collection the collection
     * @param predicate the predicate to use
     * @return first element of collection that match predicate
     */
    public static <P, T extends P> T findFirst( final Collection<T> collection,
            final Predicate<P> predicate ){
        final Collection<T> filter = Collections2.filter(collection, predicate);
        final Iterator<T> iterator = filter.iterator();
        return iterator.hasNext() ? iterator.next() : null;
    }

    /**
     * check whether collection contains at least one element that match a given predicate
     * @param <T> type of collection
     * @param collection the collection
     * @param predicate the predicate
     * @return true if collection contains element that match a given predicate, otherwise false
     */
    public static <P, T extends P> boolean exists(final Collection<T> collection,
            final Predicate<P> predicate){
        return findFirst(collection, predicate) != null;
    }
    
    
    /**
     * removes duplicates
     */
    public static <T> Collection<T> removeDuplicates(Collection<T> col){
        return MapUtils.groupBy(col, Functions.<T>identity()).keySet();
    }
    
    
    /**
     * Binds each element in a collection with defined object
     */
    public static <X,Y> Collection<Pair<Y,X>> bindLeft(final Y toBind, final Collection<X> col){
        return Collections2.transform(col, PairUtils.<X,Y>bindLeft(toBind));
    }
    
    // usage in only static way allowed
    private CollectionUtils(){
    }

}
