package org.coderu.common.utils;


import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.Validate;
import org.coderu.common.utils.types.Pair;

import com.google.common.base.Function;
import com.google.common.base.Functions;



/**
 * collection of generic utilities
 */
public final class MapUtils {

    /**
     * groups objects from a collection by a keys created by transformer
     * @param <K> type of a map key
     * @param <V> type of a map entry
     * @param col collection
     * @param value2key transformer to generate keys
     * @return a map of objects grouped by a keys created by transformer
     */
    public static <K,V> Map<K,Collection<V>> groupBy(
            final Collection<V> col, final Function<V,K> value2key){
        
        return groupBy(col,value2key, Functions.<V>identity());
    }

    /**
     * groups objects from a collection by a keys created by transformer
     * @param <K> type of a map key
     * @param <V> type of a map entry
     * @param col collection
     * @param source2key transformer to generate keys
     * @param source2value transformer to generate values
     * @exception AssertionViolation if more elements have the same key
     * @return a map of objects grouped by a keys created by transformer
     */
    private static <S,K,V> Map<K,Collection<V>> groupBy(
            final Collection<S> col, 
            final Function<S,K> source2key, 
            final Function<S,V> source2value){
        
    
        // for the sake of the reproducibility of the order of elements use LinkedHashMap
        final Map<K, Collection<V>> result = new LinkedHashMap<K, Collection<V>>();
        for(final S s: col){
            final K k = source2key.apply(s);
            final Collection<V> value = result.get(k);
            final Collection<V> list = value != null ? value : new ArrayList<V>();
            list.add(source2value.apply(s));
            result.put(k, list);
        }
        return result;
    }
    
    /**
     * maps objects from a collection by a keys created by given transformer
     * @param <K> type of a map key
     * @param <V> type of a map entry
     * @param col collection
     * @param value2key transformer to generate keys
     * @exception AssertionViolation if more elements have the same key
     * @return a map of objects from a collection by a keys created by transformer
     */
    public static <K,V> Map<K,V> mapBy(final Collection<V> col, final Function<V,K> value2key){
        return MapUtils.<V,K,V>mapBy(col, value2key, Functions.<V>identity());
    }

    /**
     * maps objects from a collection by a keys created by given transformer
     * @param <S> type of a source objects
     * @param <K> type of a map key
     * @param <V> type of a map entry
     * @param col collection
     * @param source2key transformer to generate keys
     * @param source2value transformer to generate values
     * @exception AssertionViolation if more elements have the same key
     * @return a map of objects from a collection by a keys created by transformer
     */
    private static <S,K,V> Map<K,V> mapBy(
            final Collection<S> col, 
            final Function<S,K> source2key, 
            final Function<S,V> source2value){
        
        // for the sake of the reproducibility of the order of elements use LinkedHashMap
        final Map<K,V> result = new LinkedHashMap<K,V>();
        for(final S v: col){
            final K k = source2key.apply(v);
            Validate.isTrue(!result.containsKey(k), String.format("entry with key %s exists yet", k) );
            result.put(k, source2value.apply(v));
        }
        return result;
    }
    
    /**
     * convert a map to a list of key-value pairs
     * @param <K> key type
     * @param <V> value type
     * @param map map to convert
     * @return list of key-value pairs
     */
    public static <K,V> List<Pair<K,V>> map2keyValueList(Map<K,V> map){
        final List<Pair<K,V>> result = new ArrayList<Pair<K,V>>(map.size());
        for(Map.Entry<K, V> entry : map.entrySet()){
            result.add(Pair.make(entry.getKey(), entry.getValue()));
        }
        return result;
    }

    private MapUtils(){
        // use in only static manner is allowed
    }
    
}
