package org.coderu.core.impl.explorer.parser.impl.loader.bytecode;

import static org.coderu.common.utils.CollectionUtils.*;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import org.coderu.core.api.Clazz;
import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;

import com.google.common.base.Function;


public final class ByteCodeParser extends NullParser {

    private final List<Clazz> types;
    
    public ByteCodeParser() {

        types = new LinkedList<Clazz>();
    }

    
    public List<Clazz> getTypes() {
    
        return types;
    }


    private void addName(String name) {
        if(name==null) return;
        types.add(STRING_2_CLAZZ.apply(name));
    }

    private void addNames(String[] names) {
        if(names==null) return;
        types.addAll(transform(Arrays.asList(names), STRING_2_CLAZZ));
    }


    private void addDesc(String desc) {
        addType(Type.getType(desc));
    }

    private void addType(Type t) {
        switch(t.getSort()) {
        case Type.ARRAY:
            addType(t.getElementType());
            break;
        case Type.OBJECT:
            addName(t.getClassName().replace('.','/'));
            break;
        default:
            break;
        }
    }

    private void addMethodDesc(String desc) {
        addType(Type.getReturnType(desc));
        Type[] types = Type.getArgumentTypes(desc);
        for(int i = 0; i < types.length; i++) {
            addType(types[ i]);
        }
    }

    private void addSignature(String sign) {
        addSignatureIntern(sign);
    }


    private void addTypeSignature(String sign) {
        addSignatureIntern(sign);
    }

    /**
     * @param sign  
     */
    private void addSignatureIntern(String sign) {

//        if(sign!=null) {
//            try{
//                new SignatureReader(sign).acceptType(this);
//            }catch(Exception e){
//                LOGGER.warn(e, e);
//            }
//        }
    }

    public void visit(int version, int access,
        String name, String signature,
        String superName, String[] interfaces) {

        if(signature==null) {
            addName(superName);
            addNames(interfaces);
        } else {
            addSignature(signature);
        }
    }

    public FieldVisitor visitField(int access,
        String name, String desc,
        String signature, Object value) {
        if(signature==null) {
            addDesc(desc);
        } else {
            addTypeSignature(signature);
        }
        if(value instanceof Type) {
            addType((Type) value);
        }
        return this;
    }

    public MethodVisitor visitMethod(int access,
        String name, String desc,
        String signature, String[] exceptions) {
        if(signature==null) {
            addMethodDesc(desc);
        } else {
            addSignature(signature);
        }
        addNames(exceptions);
        return this;
    }

    public AnnotationVisitor visitAnnotation(
        String desc, boolean visible) {
        addDesc(desc);
        return this;
    }

    public AnnotationVisitor
    visitParameterAnnotation(int parameter,
        String desc, boolean visible) {
        addDesc(desc);
        return this;
    }

    /**
     * Visits a type instruction
     * NEW, ANEWARRAY, CHECKCAST or INSTANCEOF.
     */
    public void visitTypeInsn(int opcode, String desc) {
        if(desc.charAt(0)=='[') {
            addDesc(desc);
        } else {
            addName(desc);
        }
    }

    /**
     * Visits a field instruction
     * GETSTATIC, PUTSTATIC, GETFIELD or PUTFIELD.
     */
    public void visitFieldInsn(int opcode,
        String owner, String name, String desc) {
        addType(Type.getObjectType(owner));
        addDesc(desc);
    }

    /**
     * Visits a method instruction INVOKEVIRTUAL,
     * INVOKESPECIAL, INVOKESTATIC or
     * INVOKEINTERFACE.
     */
    public void visitMethodInsn(int opcode,
        String owner, String name, String desc) {
        addType(Type.getObjectType(owner));
        addMethodDesc(desc);
    }

    /**
     * Visits a LDC instruction.
     */
    public void visitLdcInsn(Object cst) {
        if(cst instanceof Type) {
            addType((Type) cst);
        }
    }

    /**
     * Visits a MULTIANEWARRAY instruction.
     */
    public void visitMultiANewArrayInsn(
        String desc, int dims) {
        addDesc(desc);
    }

    /**
     * Visits a try catch block.
     */
    public void visitTryCatchBlock(Label start,
        Label end, Label handler, String type) {
        addName(type);
    }

    private static final Function<String,Clazz>  STRING_2_CLAZZ = new Function<String, Clazz>() {

        public Clazz apply(String o) {
            final String className = o.replace('/', '.');
            return new Clazz(className);
        }
    };

    //private static final Logger LOGGER = Logger.getLogger(ByteCodeParser.class);
}